# *************************************************************************
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
# *************************************************************************

DEFAULT_TASK_MEMORY = "128m"
DEFAULT_TASK_CPU = 0.25

BATCH_JOB_ARGUMENTS = (
    "rundate",
    "scheduled_time",
    "started_at",
    "batch_name",
    "job_name",
)
